@@ -1,5 +1,5 @@
 # HTTPAccess2 - HTTP accessing library.
-# Copyright (C) 2000-2004  NAKAMURA, Hiroshi  <nakahiro@sarion.co.jp>.
+# Copyright (C) 2000-2005  NAKAMURA, Hiroshi  <nakahiro@sarion.co.jp>.
 
 # This program is copyrighted free software by NAKAMURA, Hiroshi.  You can
 # redistribute it and/or modify it under the same terms of Ruby's license;
@@ -22,17 +22,11 @@ require 'http-access2/cookie'
 
 
 module HTTPAccess2
-  VERSION = '2.0'
-  RUBY_VERSION_STRING = "ruby #{ RUBY_VERSION } (#{ RUBY_RELEASE_DATE }) [#{ RUBY_PLATFORM }]"
-  s = %w$Id: http-access2.rb,v 1.39 2004/02/11 01:03:36 nahi Exp $
+  VERSION = '2.0.6'
+  RUBY_VERSION_STRING = "ruby #{RUBY_VERSION} (#{RUBY_RELEASE_DATE}) [#{RUBY_PLATFORM}]"
+  s = %w$Id: http-access2.rb 114 2005-09-13 03:20:38Z nahi $
   RCS_FILE, RCS_REVISION = s[1][/.*(?=,v$)/], s[2]
 
-  RS = "\r\n"
-  FS = "\r\n\t"
-
-  # You need to install RubyPKI.  Check
-  # http://raa.ruby-lang.org/list.rhtml?name=openssl
-  # RubyPKI requires Ruby/1.8 or later.
   SSLEnabled = begin
       require 'openssl'
       true
@@ -40,6 +34,8 @@ module HTTPAccess2
       false
     end
 
+  DEBUG_SSL = true
+
 
 # DESCRIPTION
 #   HTTPAccess2::Client -- Client to retrieve web resources via HTTP.
@@ -72,6 +68,7 @@ class Client
   attr_reader :agent_name
   attr_reader :from
   attr_reader :ssl_config
+  attr_accessor :cookie_manager
   attr_reader :test_loopback_response
 
   class << self
@@ -110,7 +107,7 @@ class Client
     @session_manager.agent_name = @agent_name
     @session_manager.from = @from
     @session_manager.ssl_config = @ssl_config
-    @cookie_manager = nil
+    @cookie_manager = WebAgent::CookieManager.new
     self.proxy = proxy
   end
 
@@ -133,6 +130,33 @@ class Client
     @session_manager.protocol_version = protocol_version
   end
 
+  def connect_timeout
+    @session_manager.connect_timeout
+  end
+
+  def connect_timeout=(connect_timeout)
+    reset_all
+    @session_manager.connect_timeout = connect_timeout
+  end
+
+  def send_timeout
+    @session_manager.send_timeout
+  end
+
+  def send_timeout=(send_timeout)
+    reset_all
+    @session_manager.send_timeout = send_timeout
+  end
+
+  def receive_timeout
+    @session_manager.receive_timeout
+  end
+
+  def receive_timeout=(receive_timeout)
+    reset_all
+    @session_manager.receive_timeout = receive_timeout
+  end
+
   def proxy
     @proxy
   end
@@ -140,10 +164,12 @@ class Client
   def proxy=(proxy)
     if proxy.nil?
       @proxy = nil
-    elsif proxy.is_a?(URI)
-      @proxy = proxy
     else
-      @proxy = URI.parse(proxy)
+      if proxy.is_a?(URI)
+        @proxy = proxy
+      else
+        @proxy = URI.parse(proxy)
+      end
       if @proxy.scheme == nil or @proxy.scheme.downcase != 'http' or
 	  @proxy.host == nil or @proxy.port == nil
 	raise ArgumentError.new("unsupported proxy `#{proxy}'")
@@ -162,6 +188,12 @@ class Client
     reset_all
   end
 
+  # if your ruby is older than 2005-09-06, do not set socket_sync = false to
+  # avoid an SSL socket blocking bug in openssl/buffering.rb.
+  def socket_sync=(socket_sync)
+    @session_manager.socket_sync = socket_sync
+  end
+
   def set_basic_auth(uri, user_id, passwd)
     unless uri.is_a?(URI)
       uri = URI.parse(uri)
@@ -170,12 +202,11 @@ class Client
   end
 
   def set_cookie_store(filename)
-    if filename
-      @cookie_manager = WebAgent::CookieManager.new(filename)
-      @cookie_manager.load_cookies
-    else
-      @cookie_manager = nil
+    if @cookie_manager.cookies_file
+      raise RuntimeError.new("overriding cookie file location")
     end
+    @cookie_manager.cookies_file = filename
+    @cookie_manager.load_cookies if filename
   end
 
   def save_cookie_store
@@ -204,20 +235,15 @@ class Client
   #   Get a_sring of message-body of response.
   #
   def get_content(uri, query = nil, extheader = {}, &block)
-    retry_number = 0
-    while retry_number < 10
-      res = get(uri, query, extheader, &block)
-      if res.status == HTTP::Status::OK
-	return res.content
-      elsif HTTP::Status.redirect?(res.status)
-	uri = @redirect_uri_callback.call(res)
-	query = nil
-	retry_number += 1
-      else
-	raise RuntimeError.new("Unexpected response: #{ res.header.inspect }")
-      end
+    retry_connect(uri, query) do |uri, query|
+      get(uri, query, extheader, &block)
+    end
+  end
+
+  def post_content(uri, body = nil, extheader = {}, &block)
+    retry_connect(uri, nil) do |uri, query|
+      post(uri, body, extheader, &block)
     end
-    raise RuntimeError.new("Retry count exceeded.")
   end
 
   def default_redirect_uri_callback(res)
@@ -308,15 +334,32 @@ class Client
   # Management interface.
 
   def reset(uri)
-    @session_manager.reset(uri) if @session_manager
+    @session_manager.reset(uri)
   end
 
   def reset_all
-    @session_manager.reset_all if @session_manager
+    @session_manager.reset_all
   end
 
 private
 
+  def retry_connect(uri, query = nil)
+    retry_number = 0
+    while retry_number < 10
+      res = yield(uri, query)
+      if res.status == HTTP::Status::OK
+	return res.content
+      elsif HTTP::Status.redirect?(res.status)
+	uri = @redirect_uri_callback.call(res)
+	query = nil
+	retry_number += 1
+      else
+	raise RuntimeError.new("Unexpected response: #{res.header.inspect}")
+      end
+    end
+    raise RuntimeError.new("Retry count exceeded.")
+  end
+
   def conn_request(conn, method, uri, query, body, extheader, &block)
     unless uri.is_a?(URI)
       uri = URI.parse(uri)
@@ -339,15 +382,23 @@ private
     if cred
       extheader << ['Authorization', "Basic " << cred]
     end
-    if @cookie_manager
-      if (cookies = @cookie_manager.find(uri))
-	extheader << ['Cookie', cookies]
-      end
+    if cookies = @cookie_manager.find(uri)
+      extheader << ['Cookie', cookies]
     end
-    req = HTTP::Message.new_request(method, uri, query, body, proxy)
+    boundary = nil
+    content_type = extheader.find { |key, value|
+      key.downcase == 'content-type'
+    }
+    if content_type && content_type[1] =~ /boundary=(.+)\z/
+      boundary = $1
+    end
+    req = HTTP::Message.new_request(method, uri, query, body, proxy, boundary)
     extheader.each do |key, value|
       req.header.set(key, value)
     end
+    if content_type.nil? and !body.nil?
+      req.header.set('content-type', 'application/x-www-form-urlencoded')
+    end
     req
   end
 
@@ -415,11 +466,11 @@ private
     res.version, res.status, res.reason = sess.get_status
     sess.get_header().each do |line|
       unless /^([^:]+)\s*:\s*(.*)$/ =~ line
-	raise RuntimeError.new("Unparsable header: '#{ line }'.") if $DEBUG
+	raise RuntimeError.new("Unparsable header: '#{line}'.") if $DEBUG
       end
       res.header.set($1, $2)
     end
-    if @cookie_manager and res.header['set-cookie']
+    if res.header['set-cookie']
       res.header['set-cookie'].each do |cookie|
 	@cookie_manager.parse(cookie, req.header.request_uri)
       end
@@ -558,40 +609,54 @@ class SSLConfig	# :nodoc:
     ctx.ciphers = @ciphers
   end
 
-  def post_connection_check(cert, sess)
-    # Check if the cert is for the host it connects.
-    unless cert
-      return false
-    end
-    cert_name = retrieve_cert_name(cert)
-    if cert_name.upcase == sess.dest.host.upcase
-      true
-    else
-      false
+  # this definition must match with the one in ext/openssl/lib/openssl/ssl.rb
+  def post_connection_check(peer_cert, hostname)
+    check_common_name = true
+    cert = peer_cert
+    cert.extensions.each{|ext|
+      next if ext.oid != "subjectAltName"
+      ext.value.split(/,\s+/).each{|general_name|
+        if /\ADNS:(.*)/ =~ general_name
+          check_common_name = false
+          reg = Regexp.escape($1).gsub(/\\\*/, "[^.]+")
+          return true if /\A#{reg}\z/i =~ hostname
+        elsif /\AIP Address:(.*)/ =~ general_name
+          check_common_name = false
+          return true if $1 == hostname
+        end
+      }
+    }
+    if check_common_name
+      cert.subject.to_a.each{|oid, value|
+        if oid == "CN" && value.casecmp(hostname) == 0
+          return true
+        end
+      }
     end
+    raise OpenSSL::SSL::SSLError, "hostname not match"
   end
 
   # Default callback for verification: only dumps error.
-  def default_verify_callback(ok, ctx)
+  def default_verify_callback(is_ok, ctx)
     if $DEBUG
-      puts "#{ ok ? 'ok' : 'ng' }: #{ ctx.current_cert.subject }"
+      puts "#{ is_ok ? 'ok' : 'ng' }: #{ctx.current_cert.subject}"
     end
-    if !ok
+    if !is_ok
       depth = ctx.error_depth
       code = ctx.error
       msg = ctx.error_string
-      STDERR.puts "at depth #{ depth } - #{ code }: #{ msg }"
+      STDERR.puts "at depth #{depth} - #{code}: #{msg}"
     end
-    ok
+    is_ok
   end
 
   # Sample callback method:  CAUTION: does not check CRL/ARL.
-  def sample_verify_callback(ok, ctx)
-    unless ok
+  def sample_verify_callback(is_ok, ctx)
+    unless is_ok
       depth = ctx.error_depth
       code = ctx.error
       msg = ctx.error_string
-      STDERR.puts "at depth #{ depth } - #{ code }: #{ msg }" if $DEBUG
+      STDERR.puts "at depth #{depth} - #{code}: #{msg}" if $DEBUG
       return false
     end
 
@@ -642,22 +707,6 @@ private
   def change_notify
     @client.reset_all
   end
-
-  def retrieve_cert_name(cert)
-    subject_alt_name = cert.extensions.find { |ex| ex.oid == 'subjectAltName' }
-    if subject_alt_name
-      subject_alt_name.value.split(/\s*,\s/).each do |alt_name_pair|
-	alt_tag, alt_name = alt_name_pair.split(/:/)
-	if alt_tag == 'DNS'
-	  return alt_name
-	end
-      end
-    end
-    if (cn = cert.subject.to_a.find { |rdn| rdn[0] == 'CN' })
-      return cn[1]
-    end
-    nil
-  end
 end
 
 
@@ -672,7 +721,7 @@ class BasicAuth	# :nodoc:
   def set(uri, user_id, passwd)
     uri = uri.clone
     uri.path = uri.path.sub(/\/[^\/]*$/, '/')
-    @auth[uri] = ["#{ user_id }:#{ passwd }"].pack('m').strip
+    @auth[uri] = ["#{user_id}:#{passwd}"].pack('m').strip
     @client.reset_all
   end
 
@@ -710,7 +759,7 @@ class Site	# :nodoc:
   end
 
   def addr
-    "#{ @scheme }://#{ @host }:#{ @port.to_s }"
+    "#{@scheme}://#{@host}:#{@port.to_s}"
   end
 
   def port=(port)
@@ -724,6 +773,14 @@ class Site	# :nodoc:
       false
     end
   end
+
+  def to_s
+    addr
+  end
+
+  def inspect
+    sprintf("#<%s:0x%x %s>", self.class.name, __id__, addr)
+  end
 end
 
 
@@ -780,6 +837,7 @@ class SessionManager	# :nodoc:
   attr_accessor :protocol_version	# Requested protocol version
   attr_accessor :chunk_size		# Chunk size for chunked request
   attr_accessor :debug_dev		# Device for dumping log for debugging
+  attr_accessor :socket_sync            # Boolean value for Socket#sync
 
   # These parameters are not used now...
   attr_accessor :connect_timeout
@@ -798,13 +856,14 @@ class SessionManager	# :nodoc:
 
     @protocol_version = nil
     @debug_dev = nil
+    @socket_sync = true
     @chunk_size = 4096
 
     @connect_timeout = 60
     @connect_retry = 1
     @send_timeout = 120
-    @receive_timeout = 60	# For each read_block_size bytes...
-    @read_block_size = 4096
+    @receive_timeout = 60	# For each read_block_size bytes
+    @read_block_size = 8192
 
     @ssl_config = nil
 
@@ -858,11 +917,12 @@ private
 
   def open(dest, proxy = nil)
     sess = nil
-    if (cached = get_cached_session(dest))
+    if cached = get_cached_session(dest)
       sess = cached
     else
       sess = Session.new(dest, @agent_name, @from)
       sess.proxy = proxy
+      sess.socket_sync = @socket_sync
       sess.requested_version = @protocol_version if @protocol_version
       sess.connect_timeout = @connect_timeout
       sess.connect_retry = @connect_retry
@@ -883,7 +943,7 @@ private
   end
 
   def close(dest)
-    if (cached = get_cached_session(dest))
+    if cached = get_cached_session(dest)
       cached.close
       true
     else
@@ -926,21 +986,34 @@ end
 # HTTPAccess2::SSLSocketWrap
 #
 class SSLSocketWrap
-  def initialize(socket, context)
+  def initialize(socket, context, debug_dev = nil)
     unless SSLEnabled
-      raise RuntimeError.new("Ruby/OpenSSL module is required for https access.")
+      raise RuntimeError.new(
+        "Ruby/OpenSSL module is required for https access.")
     end
     @context = context
     @socket = socket
     @ssl_socket = create_ssl_socket(@socket)
+    @debug_dev = debug_dev
   end
 
   def ssl_connect
     @ssl_socket.connect
   end
 
-  def post_connection_check(sess)
-    @context.post_connection_check(@ssl_socket.peer_cert, sess)
+  def post_connection_check(host)
+    verify_mode = @context.verify_mode || OpenSSL::SSL::VERIFY_NONE
+    if verify_mode == OpenSSL::SSL::VERIFY_NONE
+      return
+    elsif @ssl_socket.peer_cert.nil? and
+        check_mask(verify_mode, OpenSSL::SSL::VERIFY_FAIL_IF_NO_PEER_CERT)
+      raise OpenSSL::SSL::SSLError, "no peer cert"
+    end
+    hostname = host.host
+    if @ssl_socket.respond_to?(:post_connection_check)
+      @ssl_socket.post_connection_check(hostname)
+    end
+    @context.post_connection_check(@ssl_socket.peer_cert, hostname)
   end
 
   def peer_cert
@@ -965,19 +1038,41 @@ class SSLSocketWrap
   end
 
   def gets(*args)
-    @ssl_socket.gets(*args)
+    str = @ssl_socket.gets(*args)
+    @debug_dev << str if @debug_dev
+    str
   end
 
   def read(*args)
-    @ssl_socket.read(*args)
+    str = @ssl_socket.read(*args)
+    @debug_dev << str if @debug_dev
+    str
   end
 
   def <<(str)
-    @ssl_socket.write(str)
+    rv = @ssl_socket.write(str)
+    @debug_dev << str if @debug_dev
+    rv
+  end
+
+  def flush
+    @ssl_socket.flush
+  end
+
+  def sync
+    @ssl_socket.sync
+  end
+
+  def sync=(sync)
+    @ssl_socket.sync = sync
   end
 
 private
 
+  def check_mask(value, mask)
+    value & mask == mask
+  end
+
   def create_ssl_socket(socket)
     ssl_socket = nil
     if OpenSSL::SSL.const_defined?("SSLContext")
@@ -1000,6 +1095,7 @@ class DebugSocket < TCPSocket
 
   class << self
     def create_socket(host, port, debug_dev)
+      debug_dev << "! CONNECT TO #{host}:#{port}\n"
       socket = new(host, port)
       socket.debug_dev = debug_dev
       socket.log_connect
@@ -1025,13 +1121,13 @@ class DebugSocket < TCPSocket
 
   def gets(*args)
     str = super
-    @debug_dev << str
+    @debug_dev << str if str
     str
   end
 
   def read(*args)
     str = super
-    @debug_dev << str
+    @debug_dev << str if str
     str
   end
 
@@ -1063,6 +1159,7 @@ class Session	# :nodoc:
   attr_reader :dest			# Destination site
   attr_reader :src			# Source site
   attr_accessor :proxy			# Proxy site
+  attr_accessor :socket_sync            # Boolean value for Socket#sync
 
   attr_accessor :requested_version	# Requested protocol version
 
@@ -1081,6 +1178,7 @@ class Session	# :nodoc:
     @dest = dest
     @src = Site.new
     @proxy = nil
+    @socket_sync = true
     @requested_version = nil
 
     @debug_dev = nil
@@ -1109,11 +1207,12 @@ class Session	# :nodoc:
   # Send a request to the server
   def query(req)
     connect() if @state == :INIT
-
     begin
       timeout(@send_timeout) do
 	set_header(req)
 	req.dump(@socket)
+        # flush the IO stream as IO::sync mode is false
+        @socket.flush unless @socket_sync
       end
     rescue Errno::ECONNABORTED
       close
@@ -1136,6 +1235,7 @@ class Session	# :nodoc:
 
   def close
     unless @socket.nil?
+      @socket.flush
       @socket.close unless @socket.closed?
     end
     @state = :INIT
@@ -1230,12 +1330,12 @@ class Session	# :nodoc:
 
 private
 
-  LibNames = "(#{ RCS_FILE }/#{ RCS_REVISION }, #{ RUBY_VERSION_STRING })"
+  LibNames = "(#{RCS_FILE}/#{RCS_REVISION}, #{RUBY_VERSION_STRING})"
 
   def set_header(req)
     req.version = @requested_version if @requested_version
     if @user_agent
-      req.header.set('User-Agent', "#{ @user_agent } #{ LibNames }")
+      req.header.set('User-Agent', "#{@user_agent} #{LibNames}")
     end
     if @from
       req.header.set('From', @from)
@@ -1250,16 +1350,23 @@ private
       retry_number = 0
       timeout(@connect_timeout) do
 	@socket = create_socket(site)
-	@src.host = @socket.addr[3]
-	@src.port = @socket.addr[1]
+        begin
+          @src.host = @socket.addr[3]
+          @src.port = @socket.addr[1]
+        rescue SocketError
+          # to avoid IPSocket#addr problem on Mac OS X 10.3 + ruby-1.8.1.
+          # cf. [ruby-talk:84909], [ruby-talk:95827]
+        end
 	if @dest.scheme == 'https'
 	  @socket = create_ssl_socket(@socket)
 	  connect_ssl_proxy(@socket) if @proxy
 	  @socket.ssl_connect
-	  unless @socket.post_connection_check(self)
-	    raise OpenSSL::SSL::SSLError.new("Post connection check failed.")
-	  end
+	  @socket.post_connection_check(@dest)
 	end
+        # Use Ruby internal buffering instead of passing data immediatly
+        # to the underlying layer
+        # => we need to to call explicitely flush on the socket
+        @socket.sync = @socket_sync
       end
     rescue TimeoutError
       if @connect_retry == 0
@@ -1277,16 +1384,21 @@ private
   end
 
   def create_socket(site)
-    if @debug_dev	
-      DebugSocket.create_socket(site.host, site.port, @debug_dev)
-    else
-      TCPSocket.new(site.host, site.port)
+    begin
+      if @debug_dev	
+        DebugSocket.create_socket(site.host, site.port, @debug_dev)
+      else
+        TCPSocket.new(site.host, site.port)
+      end
+    rescue SystemCallError => e
+      e.message << " (#{site.host}, ##{site.port})"
+      raise
     end
   end
 
   # wrap socket with OpenSSL.
   def create_ssl_socket(raw_socket)
-    SSLSocketWrap.new(raw_socket, @ssl_config)
+    SSLSocketWrap.new(raw_socket, @ssl_config, (DEBUG_SSL ? @debug_dev : nil))
   end
 
   def connect_ssl_proxy(socket)
@@ -1299,7 +1411,6 @@ private
   end
 
   # Read status block.
-  StatusParseRegexp = %r(\AHTTP/(\d+\.\d+)\s+(\d+)(?:\s+(.*))?#{ RS }\z)
   def read_header
     if @state == :DATA
       get_data {}
@@ -1308,9 +1419,7 @@ private
     unless @state == :META
       raise InvalidState, 'state != :META'
     end
-
     parse_header(@socket)
-
     @content_length = nil
     @chunked = false
     @headers.each do |line|
@@ -1345,32 +1454,38 @@ private
         close
       end
     end
-
     @next_connection = false unless @content_length
-
     return [@version, @status, @reason]
   end
 
+  StatusParseRegexp = %r(\AHTTP/(\d+\.\d+)\s+(\d+)(?:\s+(.*))?\r?\n\z)
   def parse_header(socket)
     begin
       timeout(@receive_timeout) do
 	begin
-	  @status_line = socket.gets(RS)
-	  if @status_line.nil?
+          initial_line = socket.gets("\n")
+	  if initial_line.nil?
 	    raise KeepAliveDisconnected.new
 	  end
-	  StatusParseRegexp =~ @status_line
-	  unless $1
-	    raise BadResponse.new(@status_line)
-	  end
-	  @version, @status, @reason = $1, $2.to_i, $3
-	  @next_connection = HTTP.keep_alive_enabled?(@version) ? true : false
+	  if StatusParseRegexp =~ initial_line
+            @version, @status, @reason = $1, $2.to_i, $3
+            @next_connection = HTTP.keep_alive_enabled?(@version)
+          else
+            @version = '0.9'
+            @status = nil
+            @reason = nil
+            @next_connection = false
+            @readbuf = initial_line
+            break
+          end
 	  @headers = []
-	  until ((line = socket.gets(RS)) == RS)
+          while true
+            line = socket.gets("\n")
 	    unless line
 	      raise BadResponse.new('Unexpected EOF.')
 	    end
-	    line.sub!(/#{ RS }\z/, '')
+            line.sub!(/\r?\n\z/, '')
+            break if line.empty?
 	    if line.sub!(/^\t/, '')
       	      @headers[-1] << line
 	    else
@@ -1422,8 +1537,9 @@ private
     return data
   end
 
-  ChunkDelimiter = "0#{ RS }"
-  ChunkTrailer = "0#{ RS }#{ RS }"
+  RS = "\r\n"
+  ChunkDelimiter = "0#{RS}"
+  ChunkTrailer = "0#{RS}#{RS}"
   def read_body_chunked
     if @chunk_length == 0
       until (i = @readbuf.index(RS))
